#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals
import argparse


def parse_args():
    parser = argparse.ArgumentParser(
        description="Мониторинг блокировок в PostgreSQL",
    )
    parser.add_argument("--config", "-c", type=str,
                        help="Путь к конфигурационному файлу", required=True)
    parser.add_argument("--duration", "-d", type=int,
                        help="Длительность минимально допустимой блокировки в секундах", default=0)
    return parser.parse_args()



# coding: utf-8
from __future__ import unicode_literals, print_function
import os

import datetime
from io import BytesIO
from collections import namedtuple, defaultdict

import psycopg2
import pytz

sql_dir = os.path.join(os.path.dirname(__file__), 'sql')

BackendProcess = namedtuple('BackendProcess', ['pid', 'backend_start', 'xact_start', 'state_change',
                                               'query_start', 'query', 'blocked_by', 'wait_event_type',
                                               'wait_event', 'lock_mode'])

BlockingProcess = namedtuple('BlockingProcess', ['pid', 'backend_start', 'xact_start',
                                                 'query_start', 'query', 'state',
                                                 'is_blocked', 'blockers', 'locks'])

wait_event_type_verbose = defaultdict(lambda: 'неизвестный тип')
wait_event_type_verbose.update({
    'LWLockNamed': 'именованная легкая блокировка',
    'LWLockTranche': 'группа связанных легких блокировок',
    'Lock': 'тяжелая блокировка',
    'BufferPin': 'блокировка буфера данных',
})

state_verbose = defaultdict(lambda: 'неизвестный статус')
state_verbose.update({
    'active': 'выполняется',
    'idle': 'ожидание команды от клиента',
    'idle in transaction': 'находится в транзакции, ничего не выполняет',
    'idle in transaction (aborted)': 'прерывание транзакции',
    'fastpath function call': 'выполняется fast-path функция',
    'disabled': 'отключен сбор информации',
})


def read_sql_file(sql_file_path):
    full_path = os.path.join(sql_dir, sql_file_path)
    with open(full_path, 'r') as sql_file:
        return sql_file.read()


def get_file_from_text(text):
    buf = BytesIO()
    buf.write(text)
    buf.seek(0)
    return buf


def get_lock_report(pg_dsn, duration, tz=pytz.utc):
    duration = datetime.timedelta(seconds=duration)
    blocked_processes_sql = read_sql_file('get_blocked.sql')
    blocking_process_sql = read_sql_file('get_blocking.sql')
    connection = psycopg2.connect(dsn=pg_dsn)
    try:
        cursor = connection.cursor()
        try:
            cursor.execute(blocked_processes_sql)
            blocked_backends = [BackendProcess(*r) for r in cursor.fetchall()]
            blocked_backends = list(filter(lambda b: (b.state_change is not None)
                                      and (datetime.datetime.now(tz=tz) - b.state_change >= duration), blocked_backends))
        finally:
            cursor.close()
        report = 'Заблокированы процессы:\n' if blocked_backends else ''
        for i, blocked in enumerate(blocked_backends):
            report += '%3d. PID: %d;\n' \
                   '     Cоединение: %s;\n' \
                   '     Начало транзакции: %s;\n' \
                   '     Начало запроса: %s;\n' \
                   '     Запрос: %s\n' \
                   '     Длительность блокировки: %s;\n' \
                   '     Тип блокировки: %s;\n' \
                   '     Ожидаемое событие: %s;\n' \
                   '     Режим блокировки: %s;\n' \
                   % (i+1, blocked.pid,
                      blocked.backend_start.astimezone(tz),
                      blocked.xact_start.astimezone(tz),
                      blocked.query_start.astimezone(tz),
                      blocked.query,
                      datetime.datetime.now(tz=tz) - blocked.state_change,
                      wait_event_type_verbose[blocked.wait_event_type],
                      blocked.wait_event, blocked.lock_mode, )
            report += '     Кем заблокирован:\n'
            cursor = connection.cursor()
            try:
                cursor.execute(blocking_process_sql, (blocked.blocked_by, ))
                blocking_backends = [BlockingProcess(*r) for r in cursor.fetchall()]
            finally:
                cursor.close()
            for j, blocking_process in enumerate(blocking_backends):
                report += '     %3d. PID: %d;\n' \
                       '          Начало транзакции: %s;\n' \
                       '          Начало запроса: %s;\n' \
                       '          Запрос: %s\n' \
                       '          Состояние: %s;\n' \
                       '          Заблокирован: %s %s;\n' \
                       '          Выставляемые блоки: %s;\n' \
                       % (j+1, blocking_process.pid,
                          blocking_process.xact_start.astimezone(tz),
                          blocking_process.query_start.astimezone(tz),
                          blocking_process.query,
                          state_verbose[blocking_process.state],
                          'да' if blocking_process.is_blocked else 'нет',
                          blocking_process.blockers,
                          blocking_process.locks, )
        return report
    finally:
        connection.close()



# coding: utf-8
from __future__ import unicode_literals

DB_HOST = '127.0.0.1'
DB_PORT = 5432
DB_NAME = 'postgres'
DB_USER = 'postgres'
DB_PASSWORD = 'postgres'

TELEGRAM_TOKEN = ''
TELEGRAM_CHATS = []

TIMEZONE = 'Europe/Moscow'



# coding: utf-8
from setuptools import setup, find_packages

import pg_lock_monitor

setup(
    name='pg_lock_monitor',
    version=pg_lock_monitor.__version__,
    author='Ruslan Gilfanov',
    author_email='rg@informpartner.com',
    packages=find_packages(),
    package_dir={'pg_lock_monitor': 'pg_lock_monitor'},
    package_data={
        'pg_lock_monitor': ['sql/*']
    },
    install_requires=[
        'psycopg2',
        'pytz',
        'python-telegram-bot',
    ],
    entry_points={
        'console_scripts': [
            'pg_lock_monitor = pg_lock_monitor.pg_lock_monitor:main',
        ]
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, absolute_import
import os
import sys

from io import BytesIO

import pytz
import telegram
import datetime

try:
    from pg_lock_monitor import monitor, cli
except ImportError:
    import monitor
    import cli


default_timezone = 'UTC'
default_encoding = 'utf-8'

tlg_report_name_pattern = 'pg_lock_report_%s.txt'


def stdout_print(text):
    sys.stdout.write((text + '\n').encode(default_encoding))


class TelegramBot(object):
    def __init__(self, token, chats, _encoding=default_encoding):
        self.bot = telegram.Bot(token=token)
        self.chats = chats
        self.encoding = _encoding

    def alert(self, caption, text_for_file, filename):
        if sys.version_info[0] == 2:
            caption = caption.encode(self.encoding)
            filename = filename.encode(self.encoding)
        buf = BytesIO()
        buf.write(text_for_file.encode(self.encoding))
        for chat_id in self.chats:
            buf.seek(0)
            self.bot.send_document(chat_id, buf, filename, caption=caption)
        buf.close()


def import_from_path(py_path):
    try:
        from importlib.machinery import SourceFileLoader
        return SourceFileLoader("config", py_path).load_module()
    except ImportError:
        import imp
        return imp.load_source('config', py_path)


def main():
    options = cli.parse_args()
    config = options.config
    duration = options.duration
    stdout_print('Конфигурационный файл - %s' % os.path.normpath(config))
    if not os.path.exists(config):
        stdout_print('Файл %s не существует' % os.path.normpath(config))
    else:
        cfg = import_from_path(config)
        timezone = pytz.timezone(getattr(cfg, 'TIMEZONE', 'UTC'))
        stdout_print('Временная зона - %s' % timezone)
        telegram_token = getattr(cfg, 'TELEGRAM_TOKEN', None)
        telegram_chats = getattr(cfg, 'TELEGRAM_CHATS', [])
        telegram_bot = TelegramBot(telegram_token, telegram_chats) if telegram_token else None
        dsn = "host='%(db_host)s' port=%(db_port)s dbname='%(db_name)s' user='%(db_user)s'" % {
            'db_host': cfg.DB_HOST,
            'db_port': cfg.DB_PORT,
            'db_name': cfg.DB_NAME,
            'db_user': cfg.DB_USER,
        }
        stdout_print('Подключение %s' % dsn)
        dsn += 'password=%(db_password)s' % {'db_password': cfg.DB_PASSWORD}
        report = monitor.get_lock_report(dsn, duration, tz=timezone)
        stdout_print(report or 'Блокировки продолжительностью более %s не зафиксированы' % duration)
        if report:
            report_name = tlg_report_name_pattern % datetime.datetime.now(tz=timezone).strftime('%Y-%m-%dT%H-%M-%S')
            alarm_msg = 'Зафиксированы блокировки в PostgreSQL.'
            telegram_bot.alert(alarm_msg, report, report_name)


if __name__ == '__main__':
    main()
